// Copyright IBM Corp. 2016. All Rights Reserved.
// Node module: apiconnect-cli-util
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.

/********************************************************* {COPYRIGHT-TOP} ***
 * IBM Confidential
 * OCO Source Materials
 * 5725-C11, 5725-C12, 7198-8FX, 5725-F70, 5725-F71, 5725-F72, 5725-L30
 *
 * (C) Copyright IBM Corporation 2015  All Rights Reserved.
 *
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has been
 * deposited with the U.S. Copyright Office.
 ********************************************************* {COPYRIGHT-END} **/

var Bluemix = require('apiconnect-cli-login-bluemix');
var ClientRequest = require('../util/apim-request');
var Netrc = require('../util/netrc');
var apicConfig = require('apiconnect-cli-config');
var async = require('async');
var debug = require('debug')('apiml:auth');
var f = require('util').format;
var u = require('../util/util');
var _ = require('lodash');

var config = apicConfig.loadConfig();

var _apimanager = 'apimanager';
var DEFAULT_APIM_SERVER = 'us.apiconnect.ibmcloud.com';

function resolveHostname(options) {
  var serverConfig = null;
  if (options.uri) {
    serverConfig = config.parseUriValue(options.uri);
  } else {
    serverConfig = config.getOne('org') || { data: {} };
  }

  if (options.useDefaultRegion) {
    serverConfig.hostname = serverConfig.hostname || DEFAULT_APIM_SERVER;
  }

  if (!serverConfig.hostname) {
    throw new Error('The --server option is required.');
  }
  return serverConfig.hostname;
}

/**
 * @typedef BluemixStatus
 * @type Object
 * @property {string} url Bluemix login URL
 * @property {boolean} loggedIn
 * @property {number} expiration token expiration
 */

/**
 * @typedef Credentials
 * @type Object
 * @property {string} [login]
 * @property {string} [password]
 * @property {string} [token]
 * @property {BluemixStatus} [bluemixStatus]
 * @property {string} [hostname]
 */

/**
 * Get credentials for a api management server
 * @param {object} [options]
 * @param {string} [options.uri] URI to override value in config.
 * @param {boolean} [options.useDefaultRegion] If no server is available, login to us-south region.
 *
 * @return {Credentials}
 */
function getCredentials(options) {
  var hostname = resolveHostname(options);
  var bluemix = new Bluemix(hostname);
  var token = config.getOne(bluemix.tokenKey, apicConfig.TOKEN_STORE);
  var creds = Netrc.lookup(hostname);
  if (creds) {
    creds.username = creds.login;
  }

  if (token) {
    creds = _.assign({
      token: token,
      bluemixStatus: bluemix.status(),
    }, creds);
  }

  if (creds) {
    creds.hostname = hostname;
  }
  return creds;
}

/**
 * Login into an APIM management stack.
 *
 * @param username
 * @param password
 * @param host
 * @param callback {Function(err)} if no error is return then err is null - and that's mean that the login has
 *         been successfull.
 */
function login(username, password, host, callback) {
  var tasks = [];
  if (username.indexOf(_apimanager + '/') < 0) {
    username = _apimanager + '/' + username;
  }
  tasks.push(function(cb) {
    cb(null, username, password, host);
  });

  tasks.push(function(username, password, host, cb) {
    _authenticate(username, password, host, function(err, rsp, body) {
      if (err) {
        cb(err);
      } else if (rsp.statusCode === 200) {
        body = typeof body === 'string' ? JSON.parse(body) : body;
        cb(null, username, password, host);
      } else if (rsp.statusCode === 401) {
        var error = new Error(f('Status: %d: The credentials are invalid.',
          rsp.statusCode));
        cb(error);
      } else {
        err = new Error(f('Status: %d: %s',
          rsp.statusCode, body));
        cb(err);
      }
    });
  });

  tasks.push(function(username, password, host, cb) {
    _saveCredentials(username, password, host);
    cb(null);
  });

  tasks.push(function(cb) {
    debug('Done with authentication and validation.');
    cb(null);
  });

  async.waterfall(tasks, function(err) {
    if (err) {
      debug('Error' + err.toString());
      callback(err);

    } else {
      callback(null);
    }
  });

};

function logout(server, callback) {
  Netrc.deleteCredential(server);
  callback(null);
};

module.exports = {
  login: login,
  logout: logout,
  resolveHostname: resolveHostname,
  getCredentials: getCredentials,
  DEFAULT_APIM_SERVER: DEFAULT_APIM_SERVER,
};

/**
 * Authenticate user into APIC management defined by <tt>host</tt> using credentials defined by
 * <tt>username/password</tt>
 *
 * @param username
 * @param password
 * @param host
 * @param cb {Function(err, rsp, body} callback function that is invoked once authentication is completed
 */
function _authenticate(username, password, host, cb) {
  var opts = u.requestOptions(username, password, host);
  opts.url += '/v1/me';
  ClientRequest.get(opts, cb);
};
/**
 * Update the global configuration properties once user has logged in.
 *
 * @param username
 * @param password
 * @param host g
 */
function _saveCredentials(username, password, host) {
  Netrc.setCredential(host, username, password);
};
